#include <sta/drivers/MS56xx.hpp>

#include <sta/debug/assert.hpp>
#include <sta/debug/debug.hpp>
#include <sta/lang.hpp>

#include <cmath>


namespace sta 
{
    MS56xx::MS56xx(SPIDevice * device, Version version, DelayUsFunc delay, OsrLevel osr /* = OsrLevel::_1024 */)
        : device_{device},
          version_{version},
          delay_{delay},
          osr_{osr},
          intf_{Intf::SPI},
          C_{}
    {
        STA_ASSERT(device != nullptr);
    }

    MS56xx::MS56xx(I2CDevice * device, Version version, DelayUsFunc delay, OsrLevel osr /* = OsrLevel::_1024 */)
        : device_{device},
          version_{version},
          delay_{delay},
          osr_{osr},
          intf_{Intf::I2C},
          C_{}
    {
        STA_ASSERT(device != nullptr);
    }

    bool MS56xx::init()
    {
        // Reset device on start-up
        this->reset();

        this->initConstants();

        // Read the PROM e.g. constants
        this->readPROM();

        return true;
    }

    void MS56xx::delay()
    {
        delay_(osrDelay());
    }

    void MS56xx::setOsr(OsrLevel osr)
    {
        osr_ = osr;
    }

    void MS56xx::reset()
    {
        busCommand(MS56xx::Operations::RESET);
        delay_(MS56xx::RESET_DELAY);
    }

    void MS56xx::requestData(DataType type)
    {
        Operations op = type == DataType::PRESSURE ? D1_CONVERSION : D2_CONVERSION;

        // Request the ADC to read new data.
        busCommand(op + 2*this->osr_);

        delay_(osrDelay());
    }

    float MS56xx::getPressure(Unit unit /* = Unit::hPa */)
    {    
        requestData(TEMPERATURE);
        uint8_t buffer[3] = { 0x00, 0x00, 0x00 };
		busRead(MS56xx::Operations::ADC_RESULT, buffer, 3);

		uint32_t D2 = buffer[0] << 16 | buffer[1] << 8 | buffer[2];
    	float dT_ = D2 - C_[5];

        requestData(PRESSURE);
        busRead(MS56xx::Operations::ADC_RESULT, buffer, 3);
        uint32_t D1 = buffer[0] << 16 | buffer[1] << 8 | buffer[2];

        float offset =  C_[2] + dT_ * C_[4];
        float sens = C_[1] + dT_ * C_[3];

        // The pressure in Pa.
        float pressure = (D1 * sens * 4.76837158205E-7 - offset) * 3.051757813E-5;        
        
        // Convert to desired unit.
        pressure = convertPressure(pressure, unit);

        return pressure;
    }

    float MS56xx::getTemperature()
    {
		uint8_t buffer[3] = { 0x00, 0x00, 0x00 };
		busRead(MS56xx::Operations::ADC_RESULT, buffer, 3);

		uint32_t D2 = buffer[0] << 16 | buffer[1] << 8 | buffer[2];

    	float dT_ = D2 - C_[5];
    	return (2000 + dT_ * C_[6]) * 0.01;
    }

    void MS56xx::setPressureReference(float pressRef, float altRef, Unit unit /* = Unit::hPa */)
    {
        // Taken from: https://cdn-shop.adafruit.com/datasheets/BST-BMP180-DS000-09.pdf, page 17.
        sealevel_ = pressRef / (std::pow((1 - altRef / 44330), 5.255));
    }

    float MS56xx::getAltitudeEstimate()
    {
        float pressure = getPressure(Unit::hPa);

        // Taken from: https://cdn-shop.adafruit.com/datasheets/BST-BMP180-DS000-09.pdf, page 16
        float altitude = 44330 * (1 - std::pow(pressure / sealevel_, 1 / 5.255));

        return altitude;
    }

    void MS56xx::readPROM() {
        uint8_t buffer[2];

        for (size_t i = 0; i < 7; i++)
        {
            busRead(Operations::READ_PROM + i*2, buffer, 2);

            C_[i] *= uint_8BufferTouint16_t(buffer);
        }
    }

    void MS56xx::initConstants()
    {
        C_[0] = 1;

        switch (version_)
        {
        case _MS5607:
            C_[1] = 65536L;
            C_[2] = 131072;
            C_[3] = 7.8125E-3;
            C_[4] = 0.015625;
            C_[5] = 256;
            C_[6] = 1.1920928955E-7;

            break;
        case _MS5611:
            C_[1] = 32768L;
            C_[2] = 65536L;
            C_[3] = 3.90625E-3;
            C_[4] = 7.8125E-3;
            C_[5] = 256;
            C_[6] = 1.1920928955E-7;
        default:
            break;
        }
    }

    inline float MS56xx::convertPressure(float pressure, Unit unit)
    {
        switch (unit)
        {
        case Unit::mbar:
            return pressure / 100.0f;
        case Unit::hPa:
            return pressure / 100.0f;
        case Unit::Pa:
            return pressure;
        default:
            STA_UNREACHABLE();
            break;
        }
    }

    inline uint16_t MS56xx::uint_8BufferTouint16_t(uint8_t* buffer) {
        return (buffer[0] << 8) | buffer[1];
    }

    bool MS56xx::busCommand(uint8_t command)
    {
        this->device_->beginTransmission();
        this->device_->transfer(command);
        this->device_->endTransmission();

        return true;
    }

    uint32_t MS56xx::osrDelay() {
        // Delay times taken from:
        // https://www.amsys-sensor.com/downloads/notes/MS5XXX-C-code-example-for-MS56xx-MS57xx-MS58xx-AMSYS-an520e.pdf
        switch (osr_) {
        case _256:
            return 600;
        case _512:
            return 1200;
        case _1024:
            return 2300;
        case _2048:
            return 4600;
        case _4096:
            return 9100;
        default:
            STA_UNREACHABLE();
        }
    }

    bool MS56xx::busRead(uint8_t reg, uint8_t * buffer, size_t length)
    {
        this->device_->beginTransmission();
        this->device_->transfer(reg);
        this->device_->receive(buffer, length);
        this->device_->endTransmission();

        return true;
    }
}
